#include "aStar.bi"


sub navgrid.update_node_map( byref map as WorldStruct )

	redim gNode( ubound(map.tile, 1), ubound(map.tile, 2) )

	for y as integer = 0 to ubound(map.tile, 2)

		for x as integer = 0 to ubound(map.tile, 1)

			with gNode(x,y)

				.x = x
				.y = y
				.f = 0
				.g = 0
				.h = 0

				.blocked = false

				if useTerrainBias then

					select case as const map.tile(x,y).bType

						case WALL
							.blocked = true

						case GRASS
							.c = 0

						case SAND
							.c = 5

						case MUD
							.c = 35

						case WATER
							.c = 50

					end select

				else

					select case as const map.tile(x,y).bType

						case WALL
							.blocked = true

					end select

					.c = 0

				end if

			end with

		next

	next

end sub


'experimental function that blends the terrain penalty nodes together with a gaussian blur...
sub navGrid.blend_bias_map()

	for y1 as integer = 0 to maxY

		for x1 as integer = 0 to maxX

			dim as integer nodeBias, nodeCnt

			for y2 as integer = -1 to 1

				for x2 as integer = -1 to 1

					dim as integer tX = x1+x2
					dim as integer tY = y1+y2

					if tX < 0 	then continue for
					if tY < 0 	then continue for
					if tX > 		maxX then continue for
					if tY > 		maxY then continue for

					if gNode(tX,tY).blocked then continue for

					nodeCnt+=1

					nodeBias+=gNode(tX,tY).c

				next

			next

			if nodeCnt = 0 then continue for

			gNode(x1,y1).c = nodeBias/nodeCnt

		next

	next

end sub


function navGrid.get_lowF() as navnode ptr
	
	dim lowNode as navnode ptr

	for y as integer = 0 to maxY

		for x as integer = 0 to maxX

			if gNode(x,y).state = node_state.nOpen then

				if lowNode = 0 then

					lowNode = @gNode(x,y)

				else

					if( gNode(x,y).f < lowNode->f ) then

						lowNode = @gNode(x, y)

					end if

				end if

			end if

		next

	next

	return lowNode

end function


function navGrid.eval_node( byval parent as navNode ptr, byval x as integer, byval y as integer, byval gCost as integer ) as integer

	if x < 0 then return false
	if y < 0 then return false

	if x > maxX then return false
	if y > maxY then return false

	dim as navNode ptr curNode = @gNode(x, y)

	if curNode->blocked then return false

	if curNode->state = node_state.nOpen then

		if parent->g + gCost + curNode->c < curNode->g then

			curNode->state = node_state.nNone

		end if

	elseif curNode->state = node_state.nClosed then

		if parent->g + gCost + curNode->c < curNode->g then

			curNode->state = node_state.nNone

		end if

	end if

	if curNode->state = node_state.nNone then

		curNode->state = node_state.nOpen
		curNode->g = parent->g + gCost + curNode->c
		curNode->h = abs( curNode->x - endx ) * 10 + abs( curNode->y - endy ) * 10
		curNode->f = curNode->g + curNode->h
		curNode->parent = parent

	end if

	return true

end function


sub navgrid.eval_chunk( byval parent as navnode ptr, byval x as integer, byval y as integer )

	dim as integer isLeft 	= eval_node( parent, x - 1, y, 10 )

	dim as integer isRight 	= eval_node( parent, x + 1, y, 10 )

	dim as integer isUp 		= eval_node( parent, x, y - 1, 10 )

	dim as integer isDown 	= eval_node( parent, x, y + 1, 10 )

	if isUp 		and isLeft 	then eval_node( parent, x - 1, y - 1, 14 )

	if isDown 	and isLeft 	then eval_node( parent, x - 1, y + 1, 14 )

	if isUp 		and isRight then eval_node( parent, x + 1, y - 1, 14 )

	if isDown 	and isRight then eval_node( parent, x + 1, y + 1, 14 )

end sub


function navgrid.find_path( byref entity as EntityStruct, byref origin as vec2i, byref target as vec2i ) as integer

	'This actually works in reverse, so there's no need to reverse the array at the end
	startX = target.x
	startY = target.y
	endx   = origin.x
	endy   = origin.y

	dim as integer success = false

	maxX = ubound(gNode, 1)
	maxY = ubound(gNode, 2)

	if startX < 0 then return status_state.nBadStart
	if startY < 0 then return status_state.nBadStart
	if endX   < 0 then return status_state.nBadStart
	if endY   < 0 then return status_state.nBadStart

	if startX > maxX then return status_state.nBadStart
	if startY > maxY then return status_state.nBadStart
	if endX   > maxX then return status_state.nBadStart
	if endY   > maxY then return status_state.nBadStart

	if startX = endX and startY = endY 	then return status_state.nBadStart

	if gNode(endX, endY).blocked then return status_state.nBadStart

	if gNode(startX, startY).blocked then return status_state.nBadEnd


	for y as integer = 0 to ubound(gNode,2)

		for x as integer = 0 to ubound(gNode,1)

			with gNode(x,y)

				.state = node_state.nNone
				.parent = 0
				.h = 0
				.g = 0
				.f = 0

			end with

		next

	next


	dim as navnode ptr endNode = @gNode(endX, endY)
	dim as navnode ptr curNode = @gNode(startX, startY)
	curNode->state = node_state.nOpen

	do

		curNode = get_lowF()

		if curNode = 0 then

			exit do

		elseif curNode = endNode then

			success = true

			exit do

		end if

		curNode->state = node_state.nClosed

		eval_chunk( curNode, curNode->x, curNode->y )

	loop


	if success then

		dim as integer nx = curNode->x
		dim as integer ny = curNode->y
		dim as vec2i pnt
		erase entity.path.p

		do

			dim as integer lx = nx
			dim as integer ly = ny

			nx = gNode( lx, ly ).parent->x
			ny = gNode( lx, ly ).parent->y

			pnt = type( nx, ny )
			entity.path.Add(pnt)

			if nx = startX and ny = startY then

				return status_state.nSuccess

			end if

		loop

	end if

	return status_state.nIncomplete

end function
